#!/usr/bin/perl -w
# 
# Analyse client data
#

use strict;
use Cwd;
use File::Basename;
use Date::Calc;
use XML::Simple;
use Data::Dumper;
use Getopt::Mixed "nextOption";

# set the Date::Calc language
Date::Calc::Language(Date::Calc::Decode_Language('English'));

my $www_data_dir     = '/var/www/fci/';
my $settings_dir     = 'settings/';
my $client_data_dir  = 'client-data/';
my $project_data_dir = 'project-data/';
my $xml_data_dir     = 'xml-data/';
my $site_data_dir    = $settings_dir.'site-data/';

my %cfg = ( self                     => 'fci-analyse-clients.pl',
            version                  => '0.1',
	    
	    verbose                  => 0,
	    quiet                    => 0,
	    
	    'www-data-dir'           => $www_data_dir,
	    'settings-dir'           => $settings_dir,
	    'client-data-dir'        => $client_data_dir,
	    'project-data-dir'       => $project_data_dir,
	    'site-data-dir'          => $site_data_dir,
	    'xml-data-dir'           => $xml_data_dir,
	    'client-list-file'       => $xml_data_dir.'client-list.xml',

	    'client'                 => '',
          );


# Get commandline parameters
Getopt::Mixed::init("dir=s
		     client=s
                     verbose
		     quiet
		     help
		     d>dir
		     c>client
		     v>verbose
		     q>quiet
		     h>help
		    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'help') {
		print_usage(%cfg);
		exit 0;
	}
	
	$cfg{'www-data-dir'} = $value if($option eq 'dir');
	$cfg{'client'}       = $value if($option eq 'client');
	
	$cfg{verbose}    = 1      if($option eq 'verbose');
	$cfg{quiet}      = 1      if($option eq 'quiet');
}
Getopt::Mixed::cleanup();

# if quiet was requested together with verbose, disable verbose which contradicts quiet
if($cfg{quiet} && $cfg{verbose}) {
	#print "--quiet overrides --verbose\n";
	$cfg{verbose} = 0;
}

# add directory separator to the end of the www-data-dir path if missing
$cfg{'www-data-dir'} .= '/' if(substr($cfg{'www-data-dir'}, -1, 1) ne '/');

my $dir = Cwd::cwd;
if(-d $cfg{'www-data-dir'} && -x $cfg{'www-data-dir'}) {
	print "Changing working directory to: $cfg{'www-data-dir'}\n" if($cfg{verbose});
	Cwd::chdir $cfg{'www-data-dir'};
	print "\n" if($cfg{verbose});
}
else {
	print "Error: Cannot change to directory: $cfg{'www-data-dir'}\n";
	exit 1;
}


# show what's print()ed without waiting for the newline $|=1;
$|=1;


# check file existance and permissions
if(!-r $cfg{'client-list-file'}) {
	print "Error: Cannot read project-images.xml file!\n";
	exit 1;
}


# read the xml files
my $client_list    = read_xml(%cfg, Name => 'client-list')    || die "Error: Cannot read client-list XML file!\n";

#print Dumper($client_list);

if($cfg{client} eq '') {
	# analyse all clients

	foreach my $client (keys %{$client_list->{client_list}->{client}}) {	
		analyse_client($client);
		print "\n" if($cfg{verbose});
	}
}
else {
	analyse_client($cfg{client});
}

exit 0; # exit normally

################################[ Subroutines ]################################

sub get_date {
        my %option = @_;

        my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();

        $month = '0'.$month if(length($month) == 1);
        $day   = '0'.$day   if(length($day)   == 1);
        $hour  = '0'.$hour  if(length($hour)  == 1);
        $min   = '0'.$min   if(length($min)   == 1);
        $sec   = '0'.$sec   if(length($sec)   == 1);

        my $date = "$year-$month-$day";
        my $time = "$hour:$min:$sec";

        if(!$option{date} && !$option{time}) {
                return "$date $time";
        }
        elsif($option{time} && !$option{date}) {
                return $time;
        }
        elsif($option{date} && !$option{time}) {
                return $date;
        }
        else {
                return ($date, $time);
        }
}


##############
#
#     Routine: read_file()
#  Parameters: path to file
#     Returns: content of the file
# 	       undef in case of error
# Description: return the contents of a file
#

sub read_file {
	my $file    = $_[0];
	my $verbose = $_[1] ? $_[1] : 0;
	
	print "Reading file from disk: $file ... " if($verbose);
	if(-r $file) {
		my $data = '';

		open(F, "< $file") || return undef;
		while(<F>) {
			$data .= $_;
		}
		close F;

		print "Done.\n" if($verbose);

		return $data;
	}
	else {
		print "FAILED!\n" if($verbose);
		return undef;
	}
}


##############
#
#     Routine: save_file()
#  Parameters: path to file
# 	       data to save into the file
# 	       optional hash with
# 	         mode    => file premission in octal
# 	         binary  => flag to set binary mode (only needed on Windows)
# 	         verbose => flag to set verbose output
#     Returns: 1 for successfull save or undef for unsuccessfull save
#              the reason for the failure or an empty string
# Description: write a file to disk if the required permissions are set
#

sub save_file {
	my $filename = shift;
	my $content  = shift;
	my %option   = @_;

	my ($success, $reason) = (1, '');

	print "Saving file to disk: $filename ... " if($option{verbose});
	if(!-e $filename || -w $filename) {
		print "Done.\n" if($option{verbose});
		
		open(F, "> $filename") || return (0, "Error: Cannot save file: $filename ($!)\n"); 
		if($option{binary} && $option{binary} == 1) {
			print "Binmode enabled.\n" if($option{verbose});
			binmode(F);
		}
		print F $content;
		close F;

		if($option{mode} && $option{mode} =~ /^0\d{3}$/) {
			print "Set permission: $option{mode}.\n" if($option{verbose});
			chmod oct($option{mode}), $filename;
		}
		else {
			if($option{verbose}) {
				print "Ignoring invalid mode: ";
				print $option{mode};
				print "\n";
			}
		}

		return 1;
	}
	else {
		print "FAILED! (File is not writeable)\n" if($option{verbose});
		return (0, "Error: File does not exists or is unwriteable!\n");
	}
}


##############
#
#     Routine: read_xml()
#  Parameters: %args
#     Returns: a hashreference of the XML structure
# Description: read an XML file while enforcing the correct structure
#

sub read_xml {
	my %args = @_;
	
	if(!$args{Name}) {
		warn "FCI read_xml: Unknown name!\n";	
		return undef;
	}

	if($args{Name} =~ /^(client|qd|team-stats)$/) {
		if(!$args{File}) {
			warn "FCI read_xml: No file specified! ($args{Name})\n";
			return undef;
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			warn "FCI read_xml: Unknown file: $args{Name}\n";
			return undef;	
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File}) {
		warn "FCI read_xml: File does not exist: $args{File}\n";
		return undef;
	}
	if(-e $args{File} && !-r $args{File}) {
		warn "FCI read_xml: Cannot read file: $args{File}\n";
		return undef;
	}
	if(!-f $args{File}) {
		warn "FCI read_xml: Not a file: $args{File}\n";
		return undef;
	}
	
	if($args{File} && -r $args{File} && -f $args{File}) {
		if($args{Name} eq 'qd-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' },
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			return XMLin( $args{File},
			              KeyAttr    => 'team_rank',
				      ForceArray => [ 'rank_entry' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team_stats => 'team' }, 
				      ForceArray => [ 'team_stats' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' }, 
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			return XMLin( $args{File},
			              KeyAttr    => { project  => 'number', client => 'name' }, 
				      ForceArray => [ 'project', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			return XMLin( $args{File},
			              KeyAttr    => { image  => 'type',  setting  => 'id', 
				                      option => 'value', location => 'id' },
				      ForceArray => [ 'image', 'setting', 'option', 'location' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			return XMLin( $args{File},
				      KeyAttr    => { project => 'number', image => 'type' },
				      ForceArray => [ 'project', 'image' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-permissions') {
			return XMLin( $args{File},
			              KeyAttr    => { path => 'name' },
				      ForceArray => [ 'path' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'qd') {
			return XMLin( $args{File},
				      KeyAttr    => { index => 'number', server => 'number' },
				      ForceArray => [ 'index', 'server' ]
				    ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name', 'server' => 'number' },
				      ForceArray => [ 'server' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name' },
				      ForceArray => [ 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			return XMLin( $args{File},
			              KeyAttr    => { username => 'userid' },
				      ForceArray => [ 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team => 'number', username => 'userid' },
				      ForceArray => [ 'team', 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			return XMLin( $args{File},
			              KeyAttr    => { project => 'number' },
				      ForceArray => [ 'project' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'fci-summary') {
			return XMLin( $args{File},
				      KeepRoot   => 1
				    ) || die $@;
		}
		else {
			# unknown file
			return undef;
		}
	}

	return undef;
}


##############
#
#     Routine: write_xml()
#  Parameters: name of the XML file 
#              xml structure in a hash reference
#              %args
#     Returns: a hashreference of the XML structure
#     Returns: 1 on sucess
#              0 and string with reasons on errors
# Description: write an XML file while enforcing the correct structure
#

sub write_xml {
	my %args = @_;
	
	my ($success, $reason) = (1, '');
	
	if(!$args{Name}) {
		return (0, "Unknown name!\n");
	}
	
	if($args{Name} =~ /^(client|qd)$/) {
		if(!$args{File}) {
			return (0, "No file specified! ($args{Name})\n");
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			return (0, "Unknown file: $args{Name}\n");
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File} && !-w dirname($args{File}) ) {
		return (0, "Cannot write directory for: $args{Name}\n");
	}
	if(-e $args{File} && !-w $args{File}) {
		return (0, "Cannot write file: $args{Name}\n");
	}
	if(-e $args{File} && !-f $args{File}) {
		return (0, "Not a file: ($args{Name})\n");
	}
	
	
	if($success) {
		if($args{Name} eq 'qd-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => 'team_rank',
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team_stats => 'team' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project  => 'number', client => 'name' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { image  => 'type',  setting  => 'id', 
				                option => 'value', location => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project => 'number', image => 'type' },
				NoAttr     => 1,
				KeepRoot   => 1
		              ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name', server => 'number', 'fci' => 'version' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team => 'number', username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { project => 'number' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		else {
			# unknown file
			$success = 0;
			$reason .= "Unknown file: $args{Name}\n";
		}
		
		# allow group users write permissions
		chmod(0664, $args{File});
	}

	return ($success, $reason);
}

sub valid_known_project {
	my $project        = shift;
	my %cfg            = @_;

	print "Validating project: $project ... \n" if($cfg{verbose});

	my $known_projects = read_xml(%cfg, Name => 'known-projects') 
					|| return (0, "Error: Failed to read known-projects XML file!\n");
		
	if($project !~ /^\d+$/) {
		return (0, "Error: Invalid project number! ($project)\n");
	}
	
	if(!exists $known_projects->{known_projects}->{project}->{$project}) {
		return (0, "Error: Project not listed! ($project)\n");
	}

	print "Project validated.\n\n" if($cfg{verbose});

	return 1;
					
}


sub analyse_client {
	my $client_name = $_[0];

	if(!$client_name || !$client_list->{client_list}->{client}->{"$client_name"}) {
		print "Error: No client name supplied or client not listed!\n";
		return undef;
	}

	if($client_list->{client_list}->{client}->{"$client_name"}) {
		# Parse qd-output, look for deleted & ready for upload WUs in the queue
		
		my $client = $client_list->{client_list}->{client}->{"$client_name"};
		if(!$client->{files}->{qd_data}) {
			print "Warning: No qd_data for client: $client_name\n";
			return undef;
		}
		else {
			print "#" x 80 ."\n" if($cfg{verbose});
			print "Parsing qd-data XML file for client: $client_name ($client->{files}->{qd_data}) ...\n" if($cfg{verbose});
			my $qd_data = read_xml( %cfg, 
			                        Name => 'qd', 
						File => $client->{files}->{qd_data}
					      ) || die "Error: Cannot read qd-data XML file!\n";

			# check each queue index for deleted or ready for upload status
			foreach(my $index = 0; $index < 9; $index++) {
				if($cfg{verbose}) {
					print "Checking index: $index\n";
					#print Dumper($qd_data->{queue}->{index}->{$index})."\n";
				}
				
				my $queue_index = $qd_data->{queue}->{index}->{$index};
				if($queue_index->{status} eq 'deleted') {
					# report deleted WU
					print "Status 'deleted' found in index $index for client: $client_name\n";
					print "\n";
					
					my $prcg = print_index($queue_index);
					
					print "\n";
					print "-" x 80 ."\n";
					#print Dumper($qd_data->{queue}->{index}->{$index})."\n";
				}
				elsif($queue_index->{status} eq 'ready for upload') {
					# report ready for upload WU
					print "Status 'ready for upload' found in index $index for client: $client_name\n";
					print "\n";
					
					my $prcg = print_index($queue_index);
					
					print "\n";
					print "-" x 80 ."\n";
					#print Dumper($qd_data->{queue}->{index}->{$index})."\n";
				}
			}
		}
	}
}

sub print_index {
	my $queue_index = shift;

	#Project: 1496 (Run 13, Clone 48, Gen 8)
	my $prcg = "Project: $queue_index->{project_number}";
	print "Project: $queue_index->{project_number}";
	
	if(exists $queue_index->{folding}->{run} &&
	   exists $queue_index->{folding}->{clone} &&
	   exists $queue_index->{folding}->{generation}
	  ) {
		$prcg .= " (Run $queue_index->{folding}->{run},";
		$prcg .= " Clone $queue_index->{folding}->{clone},";
		$prcg .= " Gen $queue_index->{folding}->{generation})";
		
		print " (Run $queue_index->{folding}->{run},";
		print " Clone $queue_index->{folding}->{clone},";
		print " Gen $queue_index->{folding}->{generation})";
	}
	print "\n";
	
	print "\n";
	if($queue_index->{username}) {
		print "Username  : $queue_index->{username}\n";
	}
	if($queue_index->{team}) {
		print "Team      : $queue_index->{team}\n";
	}
	if($queue_index->{machine_id}) {
		print "Machine ID: $queue_index->{machine_id}\n";
	}
	if($queue_index->{id}) {
		print "ID        : $queue_index->{id}\n";
	}
	
	print "\n";
	if($queue_index->{issue_timestamp}) {
		print "Issued  : $queue_index->{issue_timestamp}\n";
	}
	if($queue_index->{begin_timestamp}) {
		print "Begun   : $queue_index->{begin_timestamp}\n";
	}
	if($queue_index->{end_timestamp}) {
		print "End     : $queue_index->{end_timestamp}\n";
	}
	if($queue_index->{expire_timestamp}) {
		print "Expires : $queue_index->{expire_timestamp}\n";
	}
	if($queue_index->{due_timestamp}) {
		print "Due     : $queue_index->{due_timestamp}\n";
	}
	# asof qd fr 050 end_in, expect_in, due_in & expire_in are called deadline
	if($queue_index->{deadline}) {
		print "Deadline: $queue_index->{deadline}\n";
	}
	if($queue_index->{expect_in}) {
		print "Deadline: $queue_index->{expect_in}\n";
	}
	if($queue_index->{due_in}) {
		print "Deadline: $queue_index->{due_in}\n";
	}
	if($queue_index->{expire_in}) {
		print "Deadline: $queue_index->{expire_in}\n";
	}
	
	print "\n";
	if($queue_index->{server}->{'1'}) {
		print "Server           : $queue_index->{server}->{'1'}->{address}\n";
	}
	if($queue_index->{server}->{'2'}) {
		print "Server           : $queue_index->{server}->{'2'}->{address}\n";
	}
	if($queue_index->{collection_server}) {
		print "Collection Server: $queue_index->{collection_server}\n";
	}
	if($queue_index->{upload_failures}) {
		print "Upload Failures  : $queue_index->{upload_failures}\n";
	}
	if($queue_index->{packet_size_limit}) {
		print "Packet Size Limit: $queue_index->{packet_size_limit}\n";
	}
	if($queue_index->{work_unit}->{size}) {
		print "Work Unit Size   : $queue_index->{work_unit}->{size}\n";
	}
	
	print "\n";
	if($queue_index->{assignment_info}) {
		#assignment info (le): Mon Aug  9 23:36:38 2004; AC1213B0
		print "Assignment Info\n";
		print "---------------\n";
		
		if($queue_index->{assignment_info}->{timestamp}) {
			print "Timestamp : $queue_index->{assignment_info}->{timestamp}\n";
		}
		if($queue_index->{assignment_info}->{checksum}) {
			print "Checksum  : $queue_index->{assignment_info}->{checksum}\n";
		}
		if($queue_index->{assignment_info}->{byte_order}) {
			print "Byte Order: $queue_index->{assignment_info}->{byte_order}\n";
		}
	}

	return $prcg;
}
						
sub print_usage {
	my %cfg = @_;
	
	print "$cfg{self} v$cfg{version}\n\n";
	print "Usage: $cfg{self} [OPTIONS]\n\n";
	print "Options:\n";
	print "-d, --dir <DIR>         Path of the FCI server directory\n";
	print "-c, --client <NAME>     Client to analyse (default: all)\n";
	print "-v, --verbose           Enable verbose output\n";
	print "-q, --quiet             Don't print any output at all, except for errors\n";
	print "-h, --help              Show this usage information\n";
}
