#!/usr/bin/perl -w

use strict;
use Cwd;
use Date::Calc;
use File::Basename;
use LWP::UserAgent;
use MIME::Lite;
use Getopt::Mixed "nextOption";
use XML::Simple;
#use Data::Dumper;

$|=1;

my %cfg = ( self                         => 'fci-jmol-missing-projects.pl',
            version                      => '0.1',

            'www-data-dir'               => '/var/www/fci/',
            'jmol-missing-projects-file' => 'xml-data/jmol-missing-projects.xml',
	    'known-projects-file'        => 'xml-data/known-projects.xml',

            'url'                        => 'http://jmol.sourceforge.net/fah/missingProjects.txt',

            'email-to'                   => 'Nicolas Vervelle <nicove@users.sourceforge.net>',
            'email-from'                 => 'Default Name <username@example.com>',
            'email-subject'              => '.xyz file for missing F@H project on JMOL website',

            download                     => 1,
            test                         => 0,
            verbose                      => 0,
          );

# get commandline parameters
Getopt::Mixed::init("dir=s
                     email-from=s
                     email-to=s
                     email-subject=s
                     no-download
                     force
                     test
                     verbose
                     help
                     d>dir
                     n>no-download
                     f>force
                     t>test
                     v>verbose
                     h>help
                    ");

my %parameter;
while (my ($option, $value) = nextOption()) {
        if($option eq 'help') {
                print_usage(%cfg);
                exit;
        }

        $cfg{'www-data-dir'}  = $value if($option eq 'dir');

	if($option eq 'email-from') {
        	$cfg{'email-from'}          = $value;
		$parameter{'email-from'}    = 1;
	}
	if($option eq 'email-to') {
        	$cfg{'email-to'}            = $value;
		$parameter{'email-to'}      = 1;
	}
	if($option eq 'email-subject') {
        	$cfg{'email-subject'}       = $value;
		$parameter{'email-subject'} = 1;
	}

        $cfg{'download'}      = 0      if($option eq 'no-download');
        $cfg{'force'}         = 1      if($option eq 'force');
        $cfg{'test'}          = 1      if($option eq 'test');
        $cfg{'verbose'}       = 1      if($option eq 'verbose');
}
Getopt::Mixed::cleanup();

# add directory separator to the end of the www-data-dir path if missing
$cfg{'www-data-dir'} .= '/' if(substr($cfg{'www-data-dir'}, -1, 1) ne '/');

# move to the www directory
my $dir = Cwd::cwd;
if(-d $cfg{'www-data-dir'} && -x $cfg{'www-data-dir'}) {
	print "Changing working directory to: $cfg{'www-data-dir'}\n" if($cfg{verbose});
	Cwd::chdir $cfg{'www-data-dir'};
	print "\n" if($cfg{verbose});
}
else {
	print "Error: Cannot change to directory: $cfg{'www-data-dir'}\n";
	exit 1;
}


# read xml files
my $jmol_missing_projects = read_xml(%cfg, Name => 'jmol-missing-projects') 
					|| die "Error: Cannot read jmol-missing-projects XML file!\n";
my $known_projects        = read_xml(%cfg, Name => 'known-projects') 
					|| die "Error: Cannot read known-projects XML file!\n";

#print Dumper($jmol_missing_projects);
#print Dumper($known_projects);


# map the keys in the XML file to the %cfg keys
my %xml2cfg = ( 'from'    => 'email-from',
                'to'      => 'email-to',
                'subject' => 'email-subject',
		'body'    => 'email-body'
              );
	      
foreach my $xml_key (keys %xml2cfg) {
	# override the default value in %cfg with that from the XML file
	# unless the parameter was given on the commandline
	
	my $cfg_key = $xml2cfg{$xml_key};
	if(!$parameter{$cfg_key}) {
		if($jmol_missing_projects->{jmol_missing_projects}->{email_settings}->{$xml_key} && 
		   $jmol_missing_projects->{jmol_missing_projects}->{email_settings}->{$xml_key} ne ''
		) {
			$cfg{$cfg_key} = $jmol_missing_projects->{jmol_missing_projects}->{email_settings}->{$xml_key};
		}
	}
}

if($jmol_missing_projects->{jmol_missing_projects}->{url} &&
   $jmol_missing_projects->{jmol_missing_projects}->{url} ne ''
) {
	$cfg{url} = $jmol_missing_projects->{jmol_missing_projects}->{url};
}


# get the Modified HTTP header value from the first line of the file
print "Looking up Last-Modified HTTP header ... " if($cfg{verbose});

my $last_modified = '';
if($jmol_missing_projects->{jmol_missing_projects}->{last_modified} &&
   $jmol_missing_projects->{jmol_missing_projects}->{last_modified} !~ /^\s*$/
) {
	$last_modified = $jmol_missing_projects->{jmol_missing_projects}->{last_modified};
        
	print "Found.\nLast-Modified: $last_modified\n" if($cfg{verbose});
}
else {
	print "Not found.\n" if($cfg{verbose});
}
print "\n" if($cfg{verbose});

# check the URL for modification, only download if we have an older version
if($last_modified !~ /^\s*$/ && $cfg{download}) {
        # download the file from the jmol website if it has been modified
        my $ua  = LWP::UserAgent->new();
           $ua->agent("$cfg{self}/$cfg{version}");

        print "Checking URL for modification: $cfg{url} ... " if($cfg{verbose});

        my $req = HTTP::Request->new( HEAD => $cfg{url},
                                      HTTP::Headers->new('If-Modified-Since' => $last_modified)
                                    );
        my $res = $ua->request($req);
        if($res->code eq '304') {
                # the file was not modified, skip download & processing
                print "Not Modified.\n" if($cfg{verbose});

                if($cfg{force}) {
                        # we will only download files if they were modified
                        # or if we're forced to

                        print "Parameter --force given, downloading anyway.\n" if($cfg{verbose});
                }
                else {
                        print "Local copy is up-to-date, skipping download.\n" if($cfg{verbose});
                        $cfg{download} = 0;
                }
        }
        else {
                print "Modified!\n" if($cfg{verbose});
        }
	print "\n" if($cfg{verbose});
}

# actually download the file
if($cfg{download} == 1) {
        my $ua  = LWP::UserAgent->new();
           $ua->agent("$cfg{self}/$cfg{version}");

        print "Downloading URL: $cfg{url} ... " if($cfg{verbose});

        my $req = HTTP::Request->new( GET => $cfg{url} );
        my $res = $ua->request($req);
        if($res->is_success) {
                print "Success\n" if($cfg{verbose});

                $last_modified = $res->header('Last-Modified');
                print "Last-Modified: $last_modified\n" if($cfg{verbose});
		$jmol_missing_projects->{jmol_missing_projects}->{last_modified} = $last_modified;
		
		# remove the missing project from the XML file
		delete $jmol_missing_projects->{jmol_missing_projects}->{missing_projects};

		# build the missing projects list from the downloaded content
		foreach(split /\r?\n/, $res->content) {
			if(/^(\d+)/) {
				my $project = $1;
				
				push @{$jmol_missing_projects->{jmol_missing_projects}->{missing_projects}->{project}}, $project;
			}
		}
		$jmol_missing_projects->{jmol_missing_projects}->{last_update} = get_date();

		#print Dumper($jmol_missing_projects);
		
		if(!$cfg{test}) {
			my ($success, $reason) = write_xml( %cfg,
	                                                    Name => 'jmol-missing-projects',
	                                                    XML  => $jmol_missing_projects
	                                                  );
			if(!$success) {
				print $reason;
				exit 1;
			}
			print "\n" if($cfg{verbose});
		}
        }
        else {
                print "Failed!\n" if($cfg{verbose});
                print "Error: Failed to retrieve URL! ($cfg{url})\n";
                print "Server returned: ".$res->code." ".$res->message."\n";

                exit 1;
        }
}

my %xyz_of;
my %sent_projects;
my $found_projects = 0;

# compile a list of the projects we've sent already
print "Compiling list of sent projects ... " if($cfg{verbose});
foreach(@{$jmol_missing_projects->{jmol_missing_projects}->{sent_projects}->{project}}) {
	if(/^(\d+)/) {
		$sent_projects{$1} = $1;
	}
}
print "Done.\n\n" if($cfg{verbose});


# check the known projects if it is a missing project
print "Checking for missing projects in known projects list.\n" if($cfg{verbose});
foreach my $missing (@{$jmol_missing_projects->{jmol_missing_projects}->{missing_projects}->{project}}) {
	# skip this project if it is listed in the sent projects list
	if($sent_projects{$missing}) {
		print " Already sent project $missing.\n" if($cfg{verbose});
		next;
	}
	
	foreach my $project (keys %{$known_projects->{known_projects}->{project}}) {
		if($project == $missing) {
			print "Found missing project: $missing\n";	
			print "File: ";
			print $known_projects->{known_projects}->{project}->{$project}->{files}->{xyz};
			print "\n";

			$xyz_of{$project}{name} = 'p'.$project;
			$xyz_of{$project}{file} = $cfg{'www-data-dir'}.$known_projects->{known_projects}->{project}->{$project}->{files}->{xyz};
			
			if($known_projects->{known_projects}->{project}->{$project}->{workunit_name} &&
			   $known_projects->{known_projects}->{project}->{$project}->{workunit_name} =~ /^p$project/
			) {
				$xyz_of{$project}{name} 
					= $known_projects->{known_projects}->{project}->{$project}->{workunit_name};
			}

			$found_projects++;
			
			last; # skip all remaining known projects
		}
	}
}

print "Found $found_projects of the missing projects.\n" if($cfg{verbose});

if($found_projects > 0) {
	print "\n" if($cfg{verbose});

	my $projects = '';
	foreach my $project (sort keys %xyz_of) {
		$projects .= " $project";
		$projects .= "\t$xyz_of{$project}{name}\n";
	}
	
        #my $email_body  = "Hi,\n\n";
	#   $email_body .= "This is an automatically generated message to send you one or more projects ";
	#   $email_body .= "which were listed in the JMOL missing projects text file.\n";
	#   $email_body .= "The missing projects file used was the one which was last modifed at: $last_modified\n";
	#   $email_body .= "The project .xyz files which were found are:\n$projects\n";
	#   $email_body .= "Regards,\n$cfg{'email-from'}\n";
	
	my $email_body = $cfg{'email-body'};
	if($email_body =~ s/#LAST_UPDATE#/$jmol_missing_projects->{jmol_missing_projects}->{last_update}/g) {}
	if($email_body =~ s/#LAST_MODIFIED#/$last_modified/g) {}
	if($email_body =~ s/#PROJECTS#/$projects/g) {}
	if($email_body =~ s/#FROM#/$cfg{'email-from'}/g) {}
	if($email_body =~ s/#TO#/$cfg{'email-to'}/g) {}
	if($email_body =~ s/#SUBJECT#/$cfg{'email-subject'}/g) {}
	   
	if($cfg{'test'}) {
		print "Not sending email to: $cfg{'email-to'}\n";
		print "Would have sent:\n\n";
		print "       From: $cfg{'email-from'}\n";
		print "         To: $cfg{'email-to'}\n";
		print "    Subject: $cfg{'email-subject'}\n";
		print "Attachments: ";
		
		my $i = 0;
		foreach my $project (sort keys %xyz_of) {
			print "             " if($i > 0);
			print 'p'. $project .".xyz\t($xyz_of{$project}{file})\n";
			$i++;
		}
		
		print "\n$email_body\n";
	}
	else {
		my $msg = MIME::Lite->new( From    => $cfg{'email-from'},
					   To      => $cfg{'email-to'},
					   Subject => $cfg{'email-subject'},
					   Type    => 'TEXT',
					   Data    => $email_body
					 );

		$jmol_missing_projects = read_xml(%cfg, Name => 'jmol-missing-projects')
							|| die "Error: Cannot read jmol-missing-projects XML file!\n";

		# attach the .xyz files
		foreach my $project (sort keys %xyz_of) {
			print "Attaching .xyz file for project $project ...\n";
			
			$msg->attach( Type        => 'text/plain',
				      Path        => $xyz_of{$project}{file},
				      Filename    => 'p'. $project .'.xyz',
				      Disposition => 'attachment'
				    );
			
			# add the project to the sent projects list
			push @{$jmol_missing_projects->{jmol_missing_projects}->{sent_projects}->{project}}, $project; 
		}
		print "Sending email to: $cfg{'email-to'} ...\n";
		$msg->send || die $msg->err;

		my ($success, $reason) = write_xml( %cfg,
                                                    Name => 'jmol-missing-projects',
                                                    XML  => $jmol_missing_projects
                                                  );
		if(!$success) {
			print $reason;
			exit 1;
		}
	}
}



################################[ Subroutines ]################################

sub get_date {
        my %option = @_;

        my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();

        $month = '0'.$month if(length($month) == 1);
        $day   = '0'.$day   if(length($day)   == 1);
        $hour  = '0'.$hour  if(length($hour)  == 1);
        $min   = '0'.$min   if(length($min)   == 1);
        $sec   = '0'.$sec   if(length($sec)   == 1);

        my $date = "$year-$month-$day";
        my $time = "$hour:$min:$sec";

        if(!$option{date} && !$option{time}) {
                return "$date $time";
        }
        elsif($option{time} && !$option{date}) {
                return $time;
        }
        elsif($option{date} && !$option{time}) {
                return $date;
        }
        else {
                return ($date, $time);
        }
}


##############
#
#     Routine: read_xml()
#  Parameters: %args
#     Returns: a hashreference of the XML structure
# Description: read an XML file while enforcing the correct structure
#

sub read_xml {
	my %args = @_;
	
	if(!$args{Name}) {
		warn "FCI read_xml: Unknown name!\n";	
		return undef;
	}

	if($args{Name} =~ /^(client|qd|team-stats)$/) {
		if(!$args{File}) {
			warn "FCI read_xml: No file specified! ($args{Name})\n";
			return undef;
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			warn "FCI read_xml: Unknown file: $args{Name}\n";
			return undef;	
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File}) {
		warn "FCI read_xml: File does not exist: $args{File}\n";
		return undef;
	}
	if(-e $args{File} && !-r $args{File}) {
		warn "FCI read_xml: Cannot read file: $args{File}\n";
		return undef;
	}
	if(!-f $args{File}) {
		warn "FCI read_xml: Not a file: $args{File}\n";
		return undef;
	}
	
	if($args{File} && -r $args{File} && -f $args{File}) {
		if($args{Name} eq 'qd-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' },
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			return XMLin( $args{File},
			              KeyAttr    => 'team_rank',
				      ForceArray => [ 'rank_entry' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team_stats => 'team' }, 
				      ForceArray => [ 'team_stats' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' }, 
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			return XMLin( $args{File},
			              KeyAttr    => { project  => 'number', client => 'name' }, 
				      ForceArray => [ 'project', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			return XMLin( $args{File},
			              KeyAttr    => { image  => 'type',  setting  => 'id', 
				                      option => 'value', location => 'id' },
				      ForceArray => [ 'image', 'setting', 'option', 'location' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			return XMLin( $args{File},
				      KeyAttr    => { project => 'number', image => 'type' },
				      ForceArray => [ 'project', 'image' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-permissions') {
			return XMLin( $args{File},
			              KeyAttr    => { path => 'name' },
				      ForceArray => [ 'path' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'qd') {
			return XMLin( $args{File},
				      KeyAttr    => { index => 'number', server => 'number' },
				      ForceArray => [ 'index', 'server' ]
				    ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name', 'server' => 'number' },
				      ForceArray => [ 'server' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name' },
				      ForceArray => [ 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			return XMLin( $args{File},
			              KeyAttr    => { username => 'userid' },
				      ForceArray => [ 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team => 'number', username => 'userid' },
				      ForceArray => [ 'team', 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			return XMLin( $args{File},
			              KeyAttr    => { project => 'number' },
				      ForceArray => [ 'project' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'fci-summary') {
			return XMLin( $args{File},
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'jmol-missing-projects') {
			return XMLin( $args{File},
				      ForceArray => [ 'project' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		else {
			# unknown file
			return undef;
		}
	}

	return undef;
}


##############
#
#     Routine: write_xml()
#  Parameters: name of the XML file 
#              xml structure in a hash reference
#              %args
#     Returns: a hashreference of the XML structure
#     Returns: 1 on sucess
#              0 and string with reasons on errors
# Description: write an XML file while enforcing the correct structure
#

sub write_xml {
	my %args = @_;
	
	my ($success, $reason) = (1, '');
	
	if(!$args{Name}) {
		return (0, "Error: Unknown name!\n");
	}
	
	if($args{Name} =~ /^(client|qd)$/) {
		if(!$args{File}) {
			return (0, "Error: No file specified! ($args{Name})\n");
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			return (0, "Error: Unknown file: $args{Name}\n");
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File} && !-w dirname($args{File}) ) {
		return (0, "Error: Cannot write directory for: $args{Name}\n");
	}
	if(-e $args{File} && !-w $args{File}) {
		return (0, "Error: Cannot write file: $args{Name}\n");
	}
	if(-e $args{File} && !-f $args{File}) {
		return (0, "Error: Not a file: ($args{Name})\n");
	}
	
	
	if($success) {
		if($args{Name} eq 'qd-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => 'team_rank',
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team_stats => 'team' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project  => 'number', client => 'name' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { image  => 'type',  setting  => 'id', 
				                option => 'value', location => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project => 'number', image => 'type' },
				NoAttr     => 1,
				KeepRoot   => 1
		              ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name', server => 'number', 'fci' => 'version' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team => 'number', username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { project => 'number' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'jmol-missing-projects') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				NoAttr     => 1,
				KeepRoot   => 1
			     ) || die $@;
		}
		else {
			# unknown file
			$success = 0;
			$reason .= "Error: Unknown file: $args{Name}\n";
		}
	}

	return ($success, $reason);
}

sub print_usage {
	my %cfg = @_;
	
	print "$cfg{self} v$cfg{version}\n\n";
	print "Usage: $cfg{self} [OPTIONS]\n\n";
	print "Options:\n";
	print "-d, --dir <DIR>               Path of the FCI server directory\n";
	print "    --email-from <STRING>     Email sender.\n";
	print "    --email-to <STRING>       Email recipient.\n";
	print "    --email-subject <STRING>  Email subject.\n";
	print "-t, --test                    Dry run, won't send email.\n";
	print "-n, --no-download             Don't download the missing projects file.\n";
	print "-f, --force                   Download missing project file even when it's uptodate.\n";
	print "-v, --verbose                 Enable verbose output\n";
	print "-h, --help                    Show this usage information\n";
}
