#!/usr/bin/perl -w
#
# Files hosted for qd:
# - http://linuxminded.xs4all.nl/software/qd-tools/source/qdinfo/qdinfo.dat
#   Data file for all Operating Systems
#   Checked hourly, updated whenever a new projects is registered or 
#   an existings projects credit has changed
#
# - http://linuxminded.xs4all.nl/software/qd-tools/binaries/linux/x86/qd
#   Linux Binary
#   Checked hourly, updated whenever a new projects is registered or
#   an existings projects credit has changed.
#   Download required when functional revision changes
#
# - http://linuxminded.xs4all.nl/software/qd-tools/binaries/win32/x86/qd.exe
#   Windows Binary
#   
# - http://linuxminded.xs4all.nl/software/qd-tools/binaries/freebsd/x86/qd
#   FreeBSD Binary
#   
# - http://linuxminded.xs4all.nl/software/qd-tools/binaries/openbsd/x86/qd
#   OpenBSD Binary
#
# - http://linuxminded.xs4all.nl/software/qd-tools/binaries/darwin/ppc/qd
#   Mac OS X Binary for PowerPC based Macs
#   
# - http://linuxminded.xs4all.nl/software/qd-tools/binaries/darwin/x86/qd
#   Mac OS X Binary for Intel based Macs
#   

use strict;
use File::Basename;
use Date::Calc;
use LWP::UserAgent;
use XML::Simple;
use Data::Dumper;
use Getopt::Mixed "nextOption";

# set the Date::Calc language
Date::Calc::Language(Date::Calc::Decode_Language('English'));

my %http_info;	# will store http status codes and response messages in case of errors

my %cfg = ( self                   => 'fci-update-qd-files.pl',
            version                => '0.1',
	    
	    verbose                => 0,
	    quiet                  => 0,
	    download               => 1,
	    force                  => 0,
	    action                 => 'all',
	    
	    'www-root'             => '/var/www/fci/',
	    'qd-files-file'        => 'xml-data/qd-files.xml',
	    'qd-data-dir'          => 'qd-data/',
	    
	    'data_file'            => 'qd-data/qdinfo.dat',
	    'bin_linux_i386'       => 'qd-data/qd',
	    'bin_win32_i386'       => 'qd-data/qd.exe',
	    'bin_freebsd_i386'     => 'qd-data/qd.freebsd',
	    'bin_openbsd_i386'     => 'qd-data/qd.openbsd',
	    'bin_darwin_ppc'       => 'qd-data/qd.darwin',
          );

my @os   = qw/linux win32 freebsd openbsd darwin/;
my $date = get_date();

# Get commandline parameters
Getopt::Mixed::init("dir=s
                     qdinfo
		     qd-linux
		     qd-win32
		     qd-freebsd
		     qd-openbsd
		     qd-darwin
                     no-download
		     force
                     verbose
		     quiet
		     help
		     d>dir
		     n>no-download
		     f>force
		     v>verbose
		     q>quiet
		     h>help
		    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'help') {
		print_usage(%cfg);
		exit;
	}
	
	if($option eq 'qdinfo') {
        	$cfg{update}{data_file}        = 1;
		$cfg{action}                   = 'individual';
	}
	if($option eq 'qd-linux') {
		$cfg{update}{bin_linux_i386}   = 1;
		$cfg{action}                   = 'individual';
	}
	if($option eq 'qd-win32') {
		$cfg{update}{bin_win32_i386}   = 1;
		$cfg{action}                   = 'individual';
	}
	if($option eq 'qd-freebsd') {
		$cfg{update}{bin_freebsd_i386} = 1;
		$cfg{action}                   = 'individual';
	}
	if($option eq 'qd-openbsd') {
		$cfg{update}{bin_openbsd_i386} = 1;
		$cfg{action}                   = 'individual';
	}
	if($option eq 'qd-darwin') {
		$cfg{update}{bin_darwin_i386}  = 1;
		$cfg{action}                   = 'individual';
	}
	
	if($option eq 'force') {
		$cfg{force}    = 1;
	}
	if($option eq 'no-download') {
		$cfg{download} = 0;
	}
	
	$cfg{'www-root'} = $value if($option eq 'dir');
	$cfg{verbose}    = 1      if($option eq 'verbose');
	$cfg{quiet}      = 1      if($option eq 'quiet');
}
Getopt::Mixed::cleanup();

# if force was requested, override no-download
if($cfg{force}) {
	$cfg{download} = 1;
}
# if quiet was requested together with verbose, disable verbose which contradicts quiet
if($cfg{quiet} && $cfg{verbose}) {
	#print "--quiet overrides --verbose\n";
	$cfg{verbose} = 0;
}
# add directory separator to the end of the www-root path if missing
if($^O eq 'MSWin32') {
	# Windows
	$cfg{'www-root'} .= '\\' if(substr($cfg{'www-root'}, -1, 1) ne '\\');
}
else {
	# One of the UNIX flavors is assumed
	$cfg{'www-root'} .= '/' if(substr($cfg{'www-root'}, -1, 1) ne '/');
}

# show what's print()ed without waiting for the newline $|=1;
$|=1;

# check file existance and permissions
if(!-w $cfg{'www-root'}.$cfg{'qd-files-file'}) {
	print "Error: Cannot write qd-files.xml!\n";
	exit;
}

# read the xml files
my $qd_files = XMLin( $cfg{'www-root'}.$cfg{'qd-files-file'},
                      KeyAttr    => { 'file' => 'id' },
		      ForceArray => [ 'file' ],
		      KeepRoot   => 1
		    ) || die $@;

#print Dumper($qd_files);


if($cfg{action} eq 'all') {
	# enable all files for update
	foreach my $id (keys %{$qd_files->{qd_files}->{file}}) {
		$cfg{update}{$id} = 1;
	}
}

# check each file and update if requested
foreach my $id (reverse sort keys %{$qd_files->{qd_files}->{file}}) {
	if($cfg{update}{$id}) {
		update_file($id, %cfg);
		print "\n" if($cfg{verbose});
	}
}


################################[ Subroutines ]################################

sub update_file {
	my $id       = shift;
	my %cfg      = @_;
	
	my $qd_files = XMLin( $cfg{'www-root'}.$cfg{'qd-files-file'},
			      KeyAttr    => { 'file' => 'id' },
			      ForceArray => [ 'file' ],
			      KeepRoot   => 1
			    ) || die $@;

	print "Updating $qd_files->{qd_files}->{file}->{$id}->{name} ($id) ...\n" if($cfg{verbose});

	if($qd_files->{qd_files}->{file}->{$id}->{url}) {
		my $url = $qd_files->{qd_files}->{file}->{$id}->{url};
	
		# create useragent object which will handle the HTTP connections
		my $ua  = LWP::UserAgent->new();
		   $ua->agent("$cfg{self}/$cfg{version}");
		
		# use the saved Modified header to only download 
		# if the file was actually changed
		if($qd_files->{qd_files}->{file}->{$id}->{last_modified}) {
			print "Checking URL for modification: $url ... " if($cfg{verbose});

			my $req = HTTP::Request->new( HEAD => $url,
			                              HTTP::Headers->new(
						        'If-Modified-Since' =>
							$qd_files->{qd_files}->{file}->{$id}->{last_modified}
						      )
						    );
			
			my $res = $ua->request($req);
			if($res->code eq '304') {
				# the file was not modified, skip download & processing
				print "Not Modified.\n" if($cfg{verbose});

				if($cfg{force}) {
					# we will only download files if they were modified
					# or if we're forced to
					
					if($cfg{verbose}) {
						print "Parameter --force given, downloading anyway.\n";
					}
				}
				else {
					if($cfg{verbose}) {
						print "$qd_files->{qd_files}->{file}->{$id}->{name} is up-to-date, skipping download.\n";
					}
					return;
				}
			}
			else {
				print "Modified!\n" if($cfg{verbose});	
			}
		}
		
		# actually download the file
		if($cfg{download} == 1) {
			print "Downloading URL     : $url ... " if($cfg{verbose});

			my $req = HTTP::Request->new( GET => $url );
			my $res = $ua->request($req);
			if($res->is_success) {
				print "Success\n" if($cfg{verbose});

				my $file = $cfg{'www-root'}.$qd_files->{qd_files}->{file}->{$id}->{path};
				
				my $last_modified = $res->header('Last-Modified');
				print "Last-Modified       : $last_modified\n" if($cfg{verbose});
				
				$qd_files->{qd_files}->{file}->{$id}->{last_modified} = $last_modified;
				
				print "Saving file to disk : $file\n" if($cfg{verbose});
				open(F, ">", $file) || die "Cannot open file: $file ($!)";
				print F $res->content;
				close F;
				
				my $version             = '';
				my $functional_revision = '';
				if($id eq 'data_file') {
					# just downloaded qdinfo.dat
				
					$version             = get_qdinfo_version($res->content);
					$functional_revision = get_qdinfo_functionalrevision($res->content);
				}
				else {
					# downloaded one of the binaries
					# download the version file, because we probably can't execute 
					# the foreign binaries to grab the info from there
					
					# since we downloaded a binary
					# we need to set the permissions correctly 
					# because the older versions fci-client.pl cannot do this
					chmod 0774, $file;

					if($qd_files->{qd_files}->{file}->{$id}->{info}->{url}) {
						my $url = $qd_files->{qd_files}->{file}->{$id}->{info}->{url};
						
						print "Downloading URL     : $url ... " if($cfg{verbose});
						
						my $req = HTTP::Request->new( GET => $url );
						my $res = $ua->request($req);
						if($res->is_success) {
							print "Success\n" if($cfg{verbose});
							
							#qd released 30 April 2006 (fr 040)
							if($res->content =~ /qd released (\d+\s+\w+\s+\d+)\s+\(fr (\d+)\)/) {
								$version             = $1;
								$functional_revision = $2;
							}
							else {
								print "Error: Cannot extract info from version file!\n";
								exit 1;
							}
						}
						else {
							print "Failed!\n" if($cfg{verbose});
							print "Error: Failed to retrieve URL! ($url)\n";
							print "Server returned: ".$res->code." ".$res->message."\n";
							
							exit 1;
						}
					}
					else {
					}
				}
				
				if($version && $version ne '') {
					print "Version             : $version\n" if($cfg{verbose});

					$qd_files->{qd_files}->{file}->{$id}->{info}->{date}                = $version;
				}
				if($functional_revision && $functional_revision ne '') {
					print "Functional Revision : $functional_revision\n" if($cfg{verbose});

					$qd_files->{qd_files}->{file}->{$id}->{info}->{functional_revision} = $functional_revision;
				}

				$qd_files->{qd_files}->{file}->{$id}->{last_update} = get_date();
				
				XMLout( $qd_files,
				        OutputFile => $cfg{'www-root'}.$cfg{'qd-files-file'},
					KeyAttr    => { 'file' => 'id' },
					NoAttr     => 1,
					KeepRoot   => 1
				      ) || die $@;

				
				print "Successfully updated the $qd_files->{qd_files}->{file}->{$id}->{name}.\n" if($cfg{verbose});
			}
			else {
				print "Failed!\n" if($cfg{verbose});
				print "Error: Failed to retrieve URL! ($url)\n";
				print "Server returned: ".$res->code." ".$res->message."\n";
				
				exit 1;
			}
		}
	}
	else {
		print "Error: No qdinfo.dat URL specified!\n";
		exit;
	}
}



##############
#
#     Routine: read_file()
#  Parameters: path to file
#     Returns: content of the file
# 	       undef in case of error
# Description: return the contents of a file
#

sub read_file {
	my $file = $_[0];
	
	print "Reading file from disk: $file ... " if($cfg{verbose});
	if(-r $file) {
		my $data = '';

		open(F, "< $file") || return undef;
		while(<F>) {
			$data .= $_;
		}
		close F;

		print "Done.\n" if($cfg{verbose});

		return $data;
	}
	else {
		print "FAILED!\n" if($cfg{verbose});
		return undef;
	}
}


##############
#
#     Routine: save_file()
#  Parameters: path to file
# 	       data to save into the file
# 	       optional hash with
# 	         mode   => file premission in octal
# 	         binary => flag to set binary mode (only needed on Windows)
#     Returns: 1 for successfull save or
# 	       undef for unsuccessfull save
# Description: write a file to disk if the required permissions are set
#

sub save_file {
	my $filename = shift;
	my $content  = shift;
	my %option   = @_;

	print "Saving file to disk: $filename ... " if($cfg{verbose});
	if(!-e $filename || -w $filename) {
		print "Done.\n" if($cfg{verbose});
		
		open(F, "> $filename") || return undef; 
		if($option{binary} && $option{binary} == 1) {
			print "Binmode enabled.\n" if($cfg{verbose});
			binmode(F);
		}
		print F $content;
		close F;

		if($option{mode} && $option{mode} =~ /^0\d{3}$/) {
			print "Set permission: $option{mode}.\n" if($cfg{verbose});
			chmod oct($option{mode}), $filename;
		}

		return 1;
	}
	else {
		print "FAILED! (File is not writeable)\n" if($cfg{verbose});
		return undef;
	}
}


##############
#
#     Routine: fetch_url()
#  Parameters: user-agent object
#              url
#     Returns: page content on success or
#              undef in case of no success
# Description: retrieve a file via an HTTP GET request
# 

sub fetch_url {
	my $ua  = $_[0];
	my $url = $_[1];
	
	my $req = HTTP::Request->new(GET => $url);
	
	print "Retrieving URL: $url ... " if($cfg{verbose});
	my $res = $ua->request($req);
	if($res->is_success) {
		print "Success\n" if($cfg{verbose});
		
		return $res->content;
	}
	else {
		print "Failed!\n" if($cfg{verbose});
		print "Error: Failed to retrieve URL! ($url)\n";
		print "HTTP Status: ".$res->code." ".$res->message."\n";
			
		return undef;
	}
}
																	

##############
#
#     Routine: valid_url()
#  Parameters: url
#     Returns: 1 on success or
#              2 in case of authentication required or
#              0 otherwise
# Description: verify the access to a file via a HTTP HEAD request and
#              detect requirement of HTTP Authentication
# 

sub valid_url {
	my $url = $_[0];
	
	my $ua  = LWP::UserAgent->new();
	   $ua->agent($cfg{useragent});
	
	my $req = HTTP::Request->new(HEAD => $url);
	my $res = $ua->request($req);
	
	# store realm and http status info in global hash
	$http_info{$url}{realm}  = $res->header('WWW-Authenticate');
	$http_info{$url}{status} = $res->code." ".$res->message;

	if($res->is_success()) {
		return 1;
	}
	elsif($res->code == 401) {
		return 2;
	}
	else {
		return 0;
	}
}

sub get_qdinfo_version {
	my $data = $_[0];
	
	# extract the date from the content
	# da 24 May 2005 (rph)
	if($data =~ /da (\d+ \w+ \d+) \(.*?\)/) {
		return $1;
	}
	else {
		return 0;
	}
}


##############
#
#     Routine: get_qdinfo_functionalrevision()
#  Parameters: qdinfo.dat content
#     Returns: functional revision
# Description: find the functional revision in qdinfo.dat
# 

sub get_qdinfo_functionalrevision {
	my $data = $_[0];

	# extract the functional revision from the content
	# fr 029
	if($data =~ /fr (\d+)/) {
		return $1;
	}
	else {
		return 0;
	}
}

sub get_date {
	my %option = @_;

	my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();
	
	$month = '0'.$month if(length($month) == 1);
	$day   = '0'.$day   if(length($day)   == 1);
	$hour  = '0'.$hour  if(length($hour)  == 1);
	$min   = '0'.$min   if(length($min)   == 1);
	$sec   = '0'.$sec   if(length($sec)   == 1);
	
	my $date = "$year-$month-$day";
	my $time = "$hour:$min:$sec";

	if(!$option{date} && !$option{time}) {
		return "$date $time";
	}
	elsif($option{time} && !$option{date}) {
		return $time;
	}
	elsif($option{date} && !$option{time}) {
		return $date;
	}
	else {
		return ($date, $time);
	}
}

sub print_usage {
	my %cfg = @_;
	
	print "$cfg{self} v$cfg{version}\n\n";
	print "Usage: $cfg{self} [OPTIONS]\n\n";
	print "Options:\n";
	print "-d, --dir <DIR>       Path of the FCI server directory\n";
	print "    --qdinfo          Download the qdinfo.dat file. (default: download all files)\n";
	print "    --qd-linux        Download the qd binary for Linux on x86.\n";
	print "    --qd-win32        Download the qd binary for Windows on x86.\n";
	print "    --qd-freebsd      Download the qd binary for FreeBSD on x86.\n";
	print "    --qd-openbsd      Download the qd binary for OpenBSD on x86.\n";
	print "    --qd-darwin       Download the qd binary for MacOS X on PowerPC.\n";
	print "-f, --force           Force downloading of file even if they are uptodate,\n";
	print "                      overrides --no-download.\n";
	print "-n, --no-download     Don't download any files, use the ones on disk\n";
	print "-v, --verbose         Enable verbose output\n";
	print "-q, --quiet           Don't print any output at all, except for errors\n";
	print "-h, --help            Show this usage information\n";
}
